/**
 * Iterates over all the invocations and receipts contained by the given
 * message regardless of their nesting.
 *
 * @param {API.AgentMessage} message
 * @param {object} [options]
 * @param {Visitor['invocation']} [options.invocation]
 * @param {Visitor['receipt']} [options.receipt]
 * @returns {Iterable<Member>}
 */
export function iterate(message: API.AgentMessage, options?: {
    invocation?: ((self: Visitor, invocation: API.Invocation) => Iterable<Member>) | undefined;
    receipt?: ((self: Visitor, receipt: API.Receipt) => Iterable<Member>) | undefined;
} | undefined): Iterable<Member>;
/**
 * @typedef {API.Variant<{invocation: API.Invocation, receipt: API.Receipt}>} Member
 * @typedef {Map<string, API.IPLDBlock>} Blocks
 *
 * @typedef {object} Visitor
 * @property {Blocks} blocks
 * @property {(self: Visitor, invocation: API.Invocation) => Iterable<Member>} invocation
 * @property {(self: Visitor, receipt: API.Receipt) => Iterable<Member>} receipt
 */
/**
 * {@link API.AgentMessage} iterator.
 */
export class Iterator {
    /**
     *
     * @param {API.AgentMessage} message
     */
    static blocks(message: API.AgentMessage): Map<string, API.Transport.Block<unknown, number, number, 1>>;
    /**
     * Iterates all embedded invocations & receipts of the given receipt using the
     * provided visitor.
     *
     * @param {Visitor} iterator
     * @param {API.Receipt} receipt
     * @returns {Iterable<Member>}
     */
    static receipt(iterator: Visitor, receipt: API.Receipt): Iterable<Member>;
    /**
     * Iterates all embedded invocations & receipts of the given invocation using the
     * provided visitor.
     *
     * @param {Visitor} iterator
     * @param {API.Invocation} invocation
     * @returns {Iterable<Member>}
     */
    static invocation(iterator: Visitor, invocation: API.Invocation): Iterable<Member>;
    /**
     * @param {API.AgentMessage} message
     * @param {object} [options]
     * @param {Visitor['invocation']} [options.invocation]
     * @param {Visitor['receipt']} [options.receipt]
     */
    constructor(message: API.AgentMessage, { invocation, receipt }?: {
        invocation?: ((self: Visitor, invocation: API.Invocation) => Iterable<Member>) | undefined;
        receipt?: ((self: Visitor, receipt: API.Receipt) => Iterable<Member>) | undefined;
    } | undefined);
    message: API.AgentMessage<any>;
    blocks: Map<string, API.Transport.Block<unknown, number, number, 1>>;
    invocation: (self: Visitor, invocation: API.Invocation) => Iterable<Member>;
    receipt: (self: Visitor, receipt: API.Receipt) => Iterable<Member>;
    [Symbol.iterator](): Generator<({
        invocation?: undefined;
    } & {
        receipt: API.Receipt;
    }) | ({
        receipt?: undefined;
    } & {
        invocation: API.Invocation;
    }) | {
        receipt: API.Receipt<{}, {}, API.Invocation<API.Capability<API.Ability, `${string}:${string}`, unknown>>, API.SigAlg>;
    }, void, undefined>;
}
export function index(message: API.AgentMessage): Iterable<API.AgentMessageIndexRecord>;
export type Member = API.Variant<{
    invocation: API.Invocation;
    receipt: API.Receipt;
}>;
export type Blocks = Map<string, API.IPLDBlock>;
export type Visitor = {
    blocks: Blocks;
    invocation: (self: Visitor, invocation: API.Invocation) => Iterable<Member>;
    receipt: (self: Visitor, receipt: API.Receipt) => Iterable<Member>;
};
import * as API from '../types.js';
import { Receipt } from '@ucanto/core';
import { Invocation } from '@ucanto/core';
//# sourceMappingURL=agent-message.d.ts.map